(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('vue')) :
  typeof define === 'function' && define.amd ? define(['exports', 'vue'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.VueComboBlocks = {}, global.vue));
})(this, (function (exports, vue) { 'use strict';

  var controls = {
    arrowUpKey: 38,
    arrowDownKey: 40,
    escKey: 27,
    enterKey: 13,
  };

  function isFn(fn) {
    return typeof fn === 'function';
  }

  function isControlledProp(props, key) {
    return props[key] !== undefined;
  }

  function hasKeyCode(keyCode, event) {
    return keyCode === event.keyCode;
  }

  function getItemIndex(index, item, items) {
    if (index !== undefined) {
      return index;
    }
    if (!items.length) {
      return -1;
    }
    return items.indexOf(item);
  }

  function hasOwnProperty(obj, prop) {
    return Object.prototype.hasOwnProperty.call(obj, prop);
  }

  function requiredProp(fnName, propName) {
    // eslint-disable-next-line no-console
    console.error(("The property \"" + propName + "\" is required in \"" + fnName + "\""));
  }

  /**
   * Returns the next index in the list of an item that is not disabled.
   *
   * @param {number} moveAmount Number of positions to move. Negative to move backwards,
   *                            positive forwards.
   * @param {number} baseIndex The initial position to move from.
   * @param {number} itemCount The total number of items.
   * @param {Function} getItemNodeFromIndex Used to check if item is disabled.
   * @param {boolean} circular Specify if navigation is circular. Default is true.
   * @returns {number} The new index. Returns baseIndex if item is not disabled.
   *                   Returns next non-disabled item otherwise. If no non-disabled found
   *                   it will return -1.
   */
  function getNextNonDisabledIndex(
    moveAmount,
    baseIndex,
    itemCount,
    getItemNodeFromIndex,
    circular
  ) {
    var currentElementNode = getItemNodeFromIndex(baseIndex);
    if (!currentElementNode || !currentElementNode.hasAttribute('disabled')) {
      return baseIndex;
    }

    if (moveAmount > 0) {
      for (var index = baseIndex + 1; index < itemCount; index += 1) {
        var itemNode = getItemNodeFromIndex(index);
        var isDisabled = itemNode.hasAttribute('disabled');
        if (!isDisabled) {
          return index;
        }
      }
    } else {
      for (var index$1 = baseIndex - 1; index$1 >= 0; index$1 -= 1) {
        if (!getItemNodeFromIndex(index$1).hasAttribute('disabled')) {
          return index$1;
        }
      }
    }

    if (circular) {
      return moveAmount > 0
        ? getNextNonDisabledIndex(1, 0, itemCount, getItemNodeFromIndex, false)
        : getNextNonDisabledIndex(
          -1,
          itemCount - 1,
          itemCount,
          getItemNodeFromIndex,
          false
        );
    }
    return -1;
  }

  function scrollToElement(element, parentElement, hoveredIndex) {
    /* eslint-disable no-param-reassign */
    var clientHeight = parentElement.clientHeight;
    var scrollTop = parentElement.scrollTop;
    var offsetHeight = element.offsetHeight;
    var offsetTop = element.offsetTop;

    if (offsetTop < scrollTop) {
      // If scrolled down past the last item, scroll all the way to the top
      if (hoveredIndex === 0) { parentElement.scrollTop = 0; }
      else { parentElement.scrollTop = offsetTop; } // scroll up
    } else {
      var offsetBottom = offsetTop + offsetHeight;
      var scrollBottom = scrollTop + clientHeight;
      if (offsetBottom > scrollBottom) {
        parentElement.scrollTop = offsetBottom - clientHeight; // scroll down
      }
    }
  /* eslint-enable no-param-reassign */
  }

  function getProps(isVue3) {
    return Object.assign({}, (isVue3 ? {
        modelValue: {
          default: null,
        },
      } : {
        value: {
          default: null,
        },
      }
      ),
      {items: {
        type: Array,
        required: true,
      },
      itemToString: {
        type: Function,
        default: function (item) { return (item ? String(item) : ''); },
      },
      stateReducer: {
        type: Function,
        default: function (s, a) { return a.changes; },
      },
      getItemId: {
        type: Function,
        default: function default$1(i) {
          return ("v-" + (this.idCounter) + "-vue-combo-blocks-item-" + i);
        },
      },
      menuId: {
        type: String,
        default: '',
      },
      inputId: {
        type: String,
        default: '',
      },
      labelId: {
        type: String,
        default: '',
      },
      scrollIntoView: {
        type: Boolean,
        default: true,
      },
      circular: {
        type: Boolean,
        default: true,
      }});
  }

  function getInitialData (item, itemToString) { return ({
    hovered: null,
    isOpen: false,
    hoveredIndex: -1,
    selectedItem: item,
    inputValue: itemToString(item),
  }); }

  var computed = {
    menuElement: function menuElement() { return document.querySelector(("#" + (this.computedMenuId))); },
    computedMenuId: function computedMenuId() { return this.menuId || ("v-" + (this.idCounter) + "-vue-combo-blocks-menu"); },
    computedInputId: function computedInputId() { return this.inputId || ("v-" + (this.idCounter) + "-vue-combo-blocks-input"); },
    computedLabelId: function computedLabelId() { return this.labelId || ("v-" + (this.idCounter) + "-vue-combo-blocks-label"); },
    selectedIndex: function selectedIndex() { return this.items.indexOf(this.selectedItem); },
  };

  var InputKeyDownArrowDown = 'InputKeyDownArrowDown';

  var InputKeyDownArrowUp = 'InputKeyDownArrowUp';
  var InputKeyDownTab = 'InputKeyDownTab';
  var InputKeyUpEscape = 'InputKeyUpEscape';
  // export const InputKeyDownHome = 'InputKeyDownHome';
  // export const InputKeyDownEnd = 'InputKeyDownEnd';
  var InputKeyUpEnter = 'InputKeyUpEnter';
  var InputChange = 'InputChange';
  var InputBlur = 'InputBlur';

  var MenuMouseLeave = 'MenuMouseLeave';

  var ItemMouseMove = 'ItemMouseMove';
  var ItemClick = 'ItemClick';

  // export const ToggleButtonClick = 'ToggleButtonClick';

  // export const FunctionToggleMenu = 'FunctionToggleMenu';
  var FunctionOpenMenu = 'FunctionOpenMenu';
  var FunctionCloseMenu = 'FunctionCloseMenu';
  // export const FunctionSetHighlightedIndex = 'FunctionSetHighlightedIndex';

  var FunctionSelectItem = 'FunctionSelectItem';
  var FunctionSetInputValue = 'FunctionSetInputValue';

  var FunctionReset = 'FunctionReset';
  var ControlledPropUpdatedSelectedItem = 'ControlledPropUpdatedSelectedItem';

  var sct = /*#__PURE__*/Object.freeze({
    __proto__: null,
    ControlledPropUpdatedSelectedItem: ControlledPropUpdatedSelectedItem,
    FunctionCloseMenu: FunctionCloseMenu,
    FunctionOpenMenu: FunctionOpenMenu,
    FunctionReset: FunctionReset,
    FunctionSelectItem: FunctionSelectItem,
    FunctionSetInputValue: FunctionSetInputValue,
    InputBlur: InputBlur,
    InputChange: InputChange,
    InputKeyDownArrowDown: InputKeyDownArrowDown,
    InputKeyDownArrowUp: InputKeyDownArrowUp,
    InputKeyDownTab: InputKeyDownTab,
    InputKeyUpEnter: InputKeyUpEnter,
    InputKeyUpEscape: InputKeyUpEscape,
    ItemClick: ItemClick,
    ItemMouseMove: ItemMouseMove,
    MenuMouseLeave: MenuMouseLeave
  });

  var inputPropsAndListeners = {
    getInputProps: function getInputProps() {
      return {
        value: this.inputValue || '',
        'aria-activedescendant': this.hovered ? this.getItemId(this.hoveredIndex) : '',
        'aria-autocomplete': 'list',
        'aria-controls': this.computedMenuId,
        'aria-labelledby': this.computedLabelId,
        id: this.computedInputId,
        autocomplete: 'off',
      };
    },
    getInputEventListeners: function getInputEventListeners(ref) {
      if ( ref === void 0 ) ref = {};
      var blur = ref.blur;
      var input = ref.input;
      var keydown = ref.keydown;
      var keyup = ref.keyup;

      return {
        blur: isFn(blur) ? blur : this.onInputBlur,
        input: isFn(input) ? input : this.onInput,
        keydown: isFn(keydown) ? keydown : this.onInputKeyDown,
        keyup: isFn(keyup) ? keyup : this.onInputKeyUp,
      };
    },
    onInputKeyDown: function onInputKeyDown(e) {
      // prevent form submit on keydown if Enter key registered in the keyup
      if (e.key === 'Enter' && this.isOpen) {
        e.preventDefault();
      // } else if (e.key === 'Tab' && this.hovered) {
      //   this.setState({
      //     inputValue: this.itemToString(this.hovered),
      //     selectedItem: this.hovered,
      //     isOpen: false,
      //   }, sct.InputKeyDownTab);
      } else if (hasKeyCode(controls.arrowDownKey, e) && !this.isOpen) {
        this.setState({ isOpen: true }, InputKeyDownArrowDown);
        this.moveSelection(e);
      } else { this.moveSelection(e); }
    },
    onInputKeyUp: function onInputKeyUp(e) {
      var enterKey = controls.enterKey;
      var escKey = controls.escKey;
      if (this.isOpen) {
        if (hasKeyCode(enterKey, e)) {
          if (this.hoveredIndex >= 0) {
            e.preventDefault();
            var itemNode = this.getItemNodeFromIndex(this.hoveredIndex);
            if (itemNode && itemNode.hasAttribute('disabled')) {
              return;
            }
            this.setState({
              inputValue: this.itemToString(this.hovered),
              selectedItem: this.hovered,
              hoveredIndex: -1,
              hovered: null,
              isOpen: false,
            }, InputKeyUpEnter);
          }
        } else if (hasKeyCode(escKey, e)) {
          this.setState({
            hoveredIndex: -1,
            hovered: null,
            isOpen: false,
          }, InputKeyUpEscape);
        }
      }
    },
    onInputBlur: function onInputBlur() {
      this.setState({
        isOpen: false,
        inputValue: this.itemToString(this.selectedItem),
        hovered: null,
        hoveredIndex: -1,
      }, InputBlur);
    },
    onInput: function onInput(e) {
      // Custom input component might return just the value
      var inputValue = e.target ? e.target.value : e;
      this.setState({
        inputValue: inputValue,
        hovered: null,
        isOpen: true,
      },
      InputChange);
    },
  };

  var itemPropsAndListeners = {
    getItemEventListeners: function getItemEventListeners(ref) {
      var this$1$1 = this;
      if ( ref === void 0 ) ref = {};
      var index = ref.index;
      var disabled = ref.disabled;
      var item = ref.item; if ( item === void 0 ) item = process.env.NODE_ENV === 'production' ? undefined
        : requiredProp('getItemEventListeners', 'item');
      var mousemove = ref.mousemove;
      var mousedown = ref.mousedown;
      var click = ref.click;

      if (disabled) { return {}; }
      return {
        mousemove: isFn(mousemove) ? mousemove : function () { return this$1$1.onItemMouseMove(index, item); },
        mousedown: isFn(mousedown) ? mousedown : this.onItemMouseDown,
        click: isFn(click) ? click : function () { return this$1$1.onItemClick(item); },
      };
    },
    onItemMouseMove: function onItemMouseMove(index, item) {
      var itemIndex = getItemIndex(index, item, this.items);
      if (this.hoveredIndex === itemIndex) { return; }
      this.setState({
        hoveredIndex: itemIndex,
        hovered: item,
      }, ItemMouseMove);
    },
    onItemMouseDown: function onItemMouseDown(e) {
      e.preventDefault();
    },
    onItemClick: function onItemClick(item) {
      // e.preventDefault();
      this.setState({
        inputValue: this.itemToString(item),
        selectedItem: item,
        isOpen: false,
        hoveredIndex: -1,
        hovered: null,
      }, ItemClick);
    },
    getItemProps: function getItemProps(ref) {
      if ( ref === void 0 ) ref = {};
      var index = ref.index;
      var disabled = ref.disabled;
      var item = ref.item; if ( item === void 0 ) item = process.env.NODE_ENV === 'production' ? undefined
        : requiredProp('getItemProps', 'item');

      var itemIndex = getItemIndex(index, item, this.items);
      var id = this.getItemId(itemIndex);
      return {
        id: id,
        disabled: disabled ? true : undefined,
        role: 'option',
        'aria-selected': this.hoveredIndex === itemIndex || this.selectedItem === item ? 'true' : 'false',
      };
    },
  };

  var menuPropsAndListeners = {
    getMenuProps: function getMenuProps() {
      return {
        id: this.computedMenuId,
        role: 'listbox',
        'aria-labelledby': this.computedLabelId,
      };
    },
    getMenuEventListeners: function getMenuEventListeners(ref) {
      if ( ref === void 0 ) ref = {};
      var mouseleave = ref.mouseleave;
      var mousedown = ref.mousedown;

      return {
        mouseleave: isFn(mouseleave) ? mouseleave : this.onMenuMouseLeave,
        mousedown: isFn(mousedown) ? mousedown : this.onMenuMouseDown,
      };
    },
    onMenuMouseLeave: function onMenuMouseLeave() {
      this.setState({
        hoveredIndex: -1,
        hovered: null,
      }, MenuMouseLeave);
    },
    onMenuMouseDown: function onMenuMouseDown(e) {
      // TODO: Prevent menu to close when clicking something
      // on a menu but not necessarily menu item.
      // This does the trick, but now you can't select and copy any text
      // in the menu.
      e.preventDefault();
    },
  };

  var methods = Object.assign({}, inputPropsAndListeners,
    itemPropsAndListeners,
    menuPropsAndListeners,
    {getComboboxProps: function getComboboxProps() {
      return {
        role: 'combobox',
        'aria-haspopup': 'listbox',
        'aria-owns': this.computedMenuId,
        'aria-expanded': this.isOpen ? 'true' : 'false',
      };
    },
    getLabelProps: function getLabelProps() {
      return {
        id: this.computedLabelId,
        for: this.computedInputId,
      };
    },
    autocompleteText: function autocompleteText(item) {
      this.setInputValue(this.itemToString(item));
    },
    reset: function reset() {
      this.setState({
        selectedItem: null,
        inputValue: '',
      }, FunctionReset);
    },
    select: function select(item) {
      this.setState({
        inputValue: this.itemToString(item),
        selectedItem: item,
        isOpen: false,
        hoveredIndex: -1,
        hovered: null,
      }, FunctionSelectItem);
    },
    setHoveredItem: function setHoveredItem(item, index, elem) {
      if (item && item !== this.hovered) {
        this.$emit('hover', item, elem);
      }
      this.hovered = item;
      this.hoveredIndex = typeof index === 'number' ? index : -1;
    },
    closeMenu: function closeMenu() {
      this.setState({
        isOpen: false,
        hoveredIndex: -1,
        hovered: null,
      }, FunctionCloseMenu);
    },
    openMenu: function openMenu() {
      this.setState({
        isOpen: true,
      }, FunctionOpenMenu);
    },
    getItemNodeFromIndex: function getItemNodeFromIndex(index) {
      return this.menuElement.querySelector(("#" + (this.getItemId(index))));
    },
    scrollItemIntoView: function scrollItemIntoView(index) {
      var itemElement = this.getItemNodeFromIndex(index);
      scrollToElement(itemElement, this.menuElement, index);
    },

    moveSelection: function moveSelection(e) {
      var this$1$1 = this;

      var itemCount = this.items.length;
      if (!this.isOpen || !itemCount) { return; }
      var isMovingDown = hasKeyCode(controls.arrowDownKey, e);
      var isMovingUp = hasKeyCode(controls.arrowUpKey, e);
      if (isMovingDown || isMovingUp) {
        e.preventDefault();
        var direction = isMovingDown * 2 - 1;
        var menuEdge = isMovingDown ? 0 : itemCount - 1;
        var hoversBetweenEdges = isMovingDown
          ? this.hoveredIndex < itemCount - 1
          : this.hoveredIndex > 0;

        var index = hoversBetweenEdges ? this.hoveredIndex + direction : menuEdge;
        var nextIndex = getNextNonDisabledIndex(
          direction,
          index,
          itemCount,
          this.getItemNodeFromIndex,
          this.circular
        );
        if (!this.circular) {
          if (isMovingUp && nextIndex > this.hoveredIndex) { return; }
          if (isMovingDown && nextIndex < this.hoveredIndex) { return; }
        }
        var item = this.items[nextIndex];

        if (this.scrollIntoView) {
          this.$nextTick(function () {
            this$1$1.scrollItemIntoView(index);
          });
        }
        this.setState({
          hoveredIndex: nextIndex,
          hovered: item,
        }, isMovingDown ? InputKeyDownArrowDown : InputKeyDownArrowUp);
      }
    },
    setInputValue: function setInputValue(value) {
      if (this.inputValue === value) {
        return;
      }
      this.setState({
        inputValue: value,
      }, FunctionSetInputValue);
    }});

  function slot(vm) {
    return {
    // prop getters
      getInputProps: vm.getInputProps,
      getItemProps: vm.getItemProps,
      getMenuProps: vm.getMenuProps,
      getComboboxProps: vm.getComboboxProps,
      getLabelProps: vm.getLabelProps,

      // event listeners
      getInputEventListeners: vm.getInputEventListeners,
      getMenuEventListeners: vm.getMenuEventListeners,
      getItemEventListeners: vm.getItemEventListeners,

      // state
      isOpen: vm.isOpen,
      selectedItem: vm.selectedItem,
      hoveredIndex: vm.hoveredIndex,
      inputValue: vm.inputValue,

      // actions
      reset: vm.reset,
      select: vm.select,
      setInputValue: vm.setInputValue,
      openMenu: vm.openMenu,
      closeMenu: vm.closeMenu,
    };
  }

  var idCounter = 0;

  function beforeCreate() {
    this.idCounter = idCounter.toString();
    idCounter += 1;
  }

  var isVue3 = !!vue.defineComponent;
  if (!isVue3 && process.env.NODE_ENV !== 'production') {
    console.error("Looks like your Vue version is 2.x, but you are using VueComboBlocks for Vue 3.\nInstall the correct version for Vue 2.x: \"npm i vue-combo-blocks@vue2\"");
  }

  var VueComboBlocks = vue.defineComponent({
    name: 'vue-combo-blocks',
    emits: {
      select: null,
      'update:modelValue': null,
      'input-value-change': null,
      'is-open-change': null,
      'hovered-index-change': null,
      'state-change': null,
      change: null,
      'show-list': null,
      focus: null,
      hover: null,

    },
    props: getProps(isVue3),
    watch: {
      modelValue: function modelValue(newValue) {
        if (isControlledProp(this.$props, 'modelValue')) {
          // Quietly set the internal state
          this.inputValue = this.itemToString(newValue);
          this.selectedItem = newValue;
        }
      },
    },
    beforeCreate: beforeCreate,
    data: function data() {
      return getInitialData(this.modelValue, this.itemToString);
    },
    computed: Object.assign({}, computed),
    methods: Object.assign({}, {setState: function setState(changes, type) {
        var oldState = this.$data;
        var newState = this.stateReducer(oldState, { changes: changes, type: type });

        var isItemSelected = hasOwnProperty(newState, 'selectedItem');
        var hasSelectedItemChanged = newState.selectedItem !== oldState.selectedItem;

        // Emit select and change events
        if (isItemSelected) {
          this.$emit('select', newState.selectedItem, type);
          // Only emit 'change' if selectedItem has changed
          if (hasSelectedItemChanged) {
            this.$emit('update:modelValue', newState.selectedItem, type);
          }
        }

        var nextFullState = {};

        // eslint-disable-next-line no-restricted-syntax
        for (var prop in newState) {
          if (hasOwnProperty(newState, prop)) {
            if (oldState[prop] !== newState[prop]) {
              nextFullState[prop] = newState[prop];
              this[prop] = nextFullState[prop];
              // Emit other events
              if (prop === 'inputValue') { this.$emit('input-value-change', newState[prop], type); }
              else if (prop === 'isOpen') { this.$emit('is-open-change', newState[prop], type); }
              else if (prop === 'hoveredIndex') { this.$emit('hovered-index-change', newState[prop], type); }
            }
          }
        }
        this.$emit('state-change', nextFullState, type);
      }},
      methods),
    render: function render() {
      return this.$slots.default(slot(this));
    },
  });

  VueComboBlocks.stateChangeTypes = sct;

  exports.default = VueComboBlocks;

  Object.defineProperty(exports, '__esModule', { value: true });

}));
